{
   imguTGA, Truevision Targa image loader for dImage
   Copyright (C) 2007. Dejan Boras

   This file is part of dImage.

   dImage is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   dImage is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with dImage.  If not, see <http://www.gnu.org/licenses/>.

   Started On:    31.05.2007.
   Last Update:   12.03.2011.
}

{$MODE OBJFPC}{$H+}{$I-}
UNIT imguTGA;

INTERFACE

   USES uError, uImage, uFile, uFileHandlers;

IMPLEMENTATION

   USES imguTGAStuff;

VAR
   tgaExt: fhTExtension;
   tgaLoader: fhTHandler;

{this routine is the purpose of this unit, it is the targa image loader}
procedure dTGALoad(data: pointer; const fhdata: fhTData);
var
   ld: imgPLoaderData;
   Header: tgaTHeader;
   imgSpec: tgaTImageSpec;

   XFileFooter: tgaTFooter;

   img: pointer;
   imgP: imgPImage;

{load an uncompressed true color image}
procedure loadUncompressedTrueColor();
begin
   {load in the image}
   imglBlockRead(ld^, img^, imgP^.Size);
   if(imgError <> 0) then exit;
end;

{load a RLE compressed image}
procedure loadRLETrueColor();
var
   {current pixel and byte}
   j, cPixel: longint;
   {one pixel}
   Pix: array[0..3] of byte = (0,0,0,0);
   bpp: longint = 0;
   chunk: longint = 0; {bytes per pixel, chunk size}
   ximg, yimg: pointer;

begin
   {NOTE: As it can be seen from this empty routine the loading of RLE compressed
   true color tga images still needs to be implemented}

   {initialize variables}
   cPixel := 0;
   ximg := img;

   {get the number of bytes per pixel(not bits, but bytes!)}
   bpp := imgP^.PixelDepth div 8;

	repeat
      {read the chunk header}
      imglBlockRead(ld^, chunk, 1);
      if(imgError <> 0) then exit;

      if(Chunk < 128) then begin {raw block}
         inc(chunk);

         {read chunk pixels into the image}
         imglBlockRead(ld^, (ximg)^, int64(chunk)*int64(bpp));
         if(imgError <> 0) then exit;

         inc(cPixel, chunk*bpp); {increment the number of processed bytes}
         inc(ximg, chunk*bpp); {increment the pointer}
      end else begin {compressed block}
         dec(chunk, 127); {get the number of pixels}
         {read the pixel}
         imglBlockRead(ld^, Pix, bpp);
         if(imgError <> 0) then exit;

         yimg := ximg;

         {fill it in chunk times}
         for j := 0 to chunk-1 do begin
            byte((yimg+0)^) := Pix[0];
            byte((yimg+1)^) := Pix[1];
            byte((yimg+2)^) := Pix[2];
            if(bpp = 4) then byte((yimg+3)^) := Pix[3];

            inc(yimg, bpp);
         end;

         inc(cPixel, chunk*bpp);{increment the number of processed bytes}
         inc(ximg, chunk*bpp); {increment the pointer}
      end;
	until(cPixel >= imgP^.Size);
end;

begin {dTGALoad}
   ld := data; imgP := ld^.img;

   {first, check for a Targa File Footer and determine if the file is in the new
   TGA format, or the old format. Not used currently but may be helpful in the
   future.}
   imglSeek(ld^, fSize(ld^.f^)-26);
   if(imgError <> 0) then exit;

   {$PUSH}{$HINTS OFF}
   {XFileFooter does not need to be initialized.}
   imglBlockRead(ld^, XFileFooter, SizeOf(tgaTFooter));{$POP}
   if(imgError <> 0) then exit;

   {if equal the file has the new format}
   if(XFileFooter.Signature = tgacFooterSignature) then;

   {HEADER}
   {read in the header}
   imglSeek(ld^, 0);
   if(imgError <> 0) then exit;
   {$PUSH}{$HINTS OFF}
   {Header does not need to be initialized.}
   imglBlockRead(ld^, Header, SizeOf(tgaTHeader));{$POP}
   if(imgError <> 0) then exit;

   {check if the image is of supported type}
   if(Header.typeImage <> TGA_UNCOMPRESSED_TRUE_COLOR) and
   (Header.typeImage <> TGA_RLE_TRUE_COLOR) then begin
      imgeRaise(eUNSUPPORTED); exit;
   end;

   {IMAGE SPECIFICATION}
   {read image specification}
   {$PUSH}{$HINTS OFF}
   {imgSpec does not need to be initialized.}
   imglBlockRead(ld^, imgSpec, SizeOf(tgaTImageSpec));{$POP}
   if(imgError <> 0) then exit;

   {load in width, height and bpp}
   imgP^.Width := imgSpec.Width;
   imgP^.Height := imgSpec.Height;
   imgP^.PixelDepth := imgSpec.PixDepth;

   {check if the values are valid and supported}
   if(imgP^.PixelDepth <> 24) and (imgP^.PixelDepth <> 32) then begin
      imgeRaise(eUNSUPPORTED); exit;
   end;
   case imgP^.PixelDepth of
      24: imgP^.PixF := PIXF_BGR;
      32: imgP^.PixF := PIXF_BGRA;
      else begin
         imgP^.PixF := PIXF_UNSUPPORTED;
         imgeRaise(eUNSUPPORTED); exit;
      end;
   end;

   {calculate size and pixels values}
   imgP^.Pixels := imgP^.Width * imgP^.Height;
   {this makes the image size come out exactly the number of bytes it actually is,
   regardless of the pixel depth}
   imgP^.Size := (imgP^.Width * imgP^.Height * imgP^.PixelDepth) div 8;

   {LOADING}
   {allocate memory for image}
   imglAllocate(ld^);
   if(imgError <> 0) then exit;

   img := imgP^.Image;

   {call the appropriate loading routine based on the type of tga image}
   if(Header.typeImage = TGA_UNCOMPRESSED_TRUE_COLOR) then loadUncompressedTrueColor()
   else if(Header.typeImage = TGA_RLE_TRUE_COLOR) then loadRLETrueColor();

   {NOTE: The color channels are always BGR or BGRA. They are not swapped.}

   imgP^.Origin := 0;
   {flip the image to make the origin top-left}
   {is origin bottom?}
   if(imgSpec.imgDescriptor and TGA_ORIGIN_VERTICAL = 0) then
      imgP^.Origin := imgP^.Origin or imgcBT_ORIGIN_VERTICAL;
   {is origin right?}
   if(imgSpec.imgDescriptor and TGA_ORIGIN_HORIZONTAL > 0) then
      imgP^.Origin := imgP^.Origin or imgcBT_ORIGIN_HORIZONTAL;

   {the image has been loaded successfully | mission accomplished}
end;

BEGIN
   {setup the extension}
   tgaExt.ext := '.tga';
   tgaExt.Handler := @tgaLoader;

   {setup the loader}
   tgaLoader.Name := 'TGA';
   tgaLoader.Handle := @dTGALoad;

   {register the extension and the loader}
   fhRegisterExt(@tgaExt, imgLoadersInfo);
   fhRegisterHandler(@tgaLoader, imgLoadersInfo);
END.
